"""
Test utilities.

Since pytest discourages putting __init__.py into test directory (i.e. making tests a package)
one cannot import from anywhere under tests folder. However, some utility classes/methods might be useful
in multiple test modules (i.e. factoryboy factories, base test classes). So this package is the place to put them.
"""
from unittest.mock import patch

from edx_ace import policy


class StubPolicy(policy.Policy):
    """ Short term policy. """
    def __init__(self, deny_value):
        self.deny_value = frozenset(deny_value)

    def check(self, message):
        return policy.PolicyResult(deny=self.deny_value)


def patch_policies(test_case, policies):
    """
    Set active policies for the duration of a test.

    Arguments:
        test_case (:class:`unittest.TestCase`): The test case that is running
        policies: The set of active policies to return from :func:`edx_ace.policy.policies`
    """
    patcher = patch(
        'edx_ace.policy.policies',
        return_value=policies
    )
    patcher.start()
    test_case.addCleanup(patcher.stop)



"""
:mod:`edx_ace.utils.date` contains utility functions used for
serializing and deserializing dates. It is intended for internal ACE
use.
"""
from datetime import datetime

from dateutil.parser import parse
from dateutil.tz import tzutc


def get_current_time():
    """The current time in the UTC timezone as a timezone-aware datetime object."""
    return datetime.now(tzutc())


def serialize(timestamp_obj):
    """
    Serialize a datetime object to an ISO8601 formatted string.

    Args:
        timestamp_obj (datetime): The timestamp to serialize.

    Returns:
        basestring: A string representation of the timestamp in ISO8601 format.
    """
    if timestamp_obj is None:
        return None

    # TODO(later): my understanding is that type checks like this are not pythonic.
    assert isinstance(timestamp_obj, datetime)

    # TODO(later): what if utcoffset() returns 0?
    if timestamp_obj.tzinfo is not None and timestamp_obj.utcoffset() is None:
        return timestamp_obj.isoformat() + 'Z'
    else:
        return timestamp_obj.isoformat()


def deserialize(timestamp_iso8601_str):
    """
    Deserialize a datetime object from an ISO8601 formatted string.

    Args:
        timestamp_iso8601_str (basestring): A timestamp as an ISO8601 formatted string.

    Returns:
        datetime: A timezone-aware python datetime object.
    """
    if timestamp_iso8601_str is None:
        return None
    return parse(timestamp_iso8601_str, yearfirst=True)



"""Unit tests for braze.py"""
from unittest.mock import Mock, patch

import ddt
import requests

from django.test import TestCase, override_settings

from edx_ace.channel.braze import BrazeEmailChannel
from edx_ace.errors import FatalChannelDeliveryError, RecoverableChannelDeliveryError
from edx_ace.message import Message
from edx_ace.presentation import render
from edx_ace.recipient import Recipient


@ddt.ddt
@override_settings(
    ACE_CHANNEL_BRAZE_API_KEY='test-api-key',
    ACE_CHANNEL_BRAZE_APP_ID='test-app-id',
    ACE_CHANNEL_BRAZE_REST_ENDPOINT='rest.braze.com',
    ACE_DEFAULT_EXPIRATION_DELAY=0,
)
class TestBrazeChannel(TestCase):
    """Tests for the braze channel"""

    def setUp(self):
        self.channel = BrazeEmailChannel()

    def deliver_email(self, lms_user_id=123, options=None, response_code=200, response_message='Success!'):
        """Sets up all the mocks for a single email"""
        message = Message(
            app_label='testapp',
            name='testmessage',
            options=options or {},
            recipient=Recipient(lms_user_id=lms_user_id, email_address='mr@robot.io'),
        )
        rendered_message = render(self.channel, message)

        with patch('edx_ace.channel.braze.requests.post') as mock_post:
            mock_response = Mock()
            mock_response.status_code = response_code
            mock_response.json.return_value = {'message': response_message, 'dispatch_id': 'test-dispatch-id'}
            if response_code >= 400:
                mock_response.raise_for_status.side_effect = requests.exceptions.HTTPError(response=mock_response)
            mock_post.return_value = mock_response
            self.channel.deliver(message, rendered_message)  # direct, so we can catch interesting exceptions ourselves

        return mock_post

    def test_happy_path(self):
        """Basic email send, no special settings"""
        mock_post = self.deliver_email()

        assert mock_post.call_count == 1
        assert mock_post.call_args[0] == ('https://rest.braze.com/messages/send',)
        assert mock_post.call_args[1] == {
            'headers': {'Authorization': 'Bearer test-api-key'},
            'json': {
                'external_user_ids': ['123'],
                'recipient_subscription_state': 'subscribed',
                'campaign_id': None,
                'messages': {
                    'email': {
                        'app_id': 'test-app-id',
                        'subject': 'template subject.txt',
                        'from': 'webmaster@localhost',
                        'reply_to': None,
                        'body': """<!DOCTYPE html>
<html>
  <head>
    template head.html\n
  </head>
  <body>
    template body.html\n\n\n\n
<tr>
    <!-- Actions -->
    <td style="padding-bottom: 20px;">
        
            <p>
                <a href="{{${set_user_to_unsubscribed_url}}}" style="color: #960909">
                    <font color="#960909"><b>Unsubscribe from this list</b></font>
                </a>
            </p>
        
    </td>
</tr>\n\n
  </body>
</html>""",  # noqa
                        'plaintext_body': 'template body.txt',
                        'message_variation_id': None,
                        'should_inline_css': False,
                    },
                },
            },
        }

    @ddt.data(
        (None, None, None),
        ('', None, None),
        ('campaign_id', 'campaign_id', None),
        ('campaign_id:variation_id', 'campaign_id', 'variation_id'),
    )
    @ddt.unpack
    def test_campaigns(self, campaign_setting, campaign_id, variation_id):
        """If set, we should pass on the campaign identifiers"""
        full_settings = {
            'testmessage': campaign_setting,
        }
        with override_settings(ACE_CHANNEL_BRAZE_CAMPAIGNS=full_settings):
            mock_post = self.deliver_email()
        assert mock_post.call_args[1]['json']['campaign_id'] == campaign_id
        assert mock_post.call_args[1]['json']['messages']['email']['message_variation_id'] == variation_id

    def test_transactional(self):
        """Transactional emails have different subscriber settings"""
        mock_post = self.deliver_email(options={'transactional': True})
        assert mock_post.call_args[1]['json']['recipient_subscription_state'] == 'all'

    def test_from_address_in_message(self):
        """Can set a from address in message options"""
        mock_post = self.deliver_email(options={'from_address': 'testing@example.com'})
        assert mock_post.call_args[1]['json']['messages']['email']['from'] == 'testing@example.com'

    @override_settings(ACE_CHANNEL_BRAZE_FROM_EMAIL='settings@example.com')
    def test_from_address_in_settings(self):
        """Can set a from address in settings, which will be preferred over the message options"""
        mock_post = self.deliver_email(options={'from_address': 'message@example.com'})
        assert mock_post.call_args[1]['json']['messages']['email']['from'] == 'settings@example.com'

    def test_reply_to(self):
        """Can set a reply-to address in options"""
        mock_post = self.deliver_email(options={'reply_to': 'reply@example.com'})
        assert mock_post.call_args[1]['json']['messages']['email']['reply_to'] == 'reply@example.com'

    @ddt.data(0, None)
    def test_lms_user_id_fallback(self, user_id):
        """Use django instead if we can't find user id"""
        with patch('edx_ace.channel.braze.DjangoEmailChannel') as mock_django:
            mock_post = self.deliver_email(lms_user_id=user_id)

        assert mock_post.call_count == 0
        assert mock_django.call_count == 1

    @ddt.data(
        (400, FatalChannelDeliveryError),
        (429, RecoverableChannelDeliveryError),
        (500, RecoverableChannelDeliveryError),
    )
    @ddt.unpack
    def test_status_raises(self, code, exception):
        with self.assertRaisesRegex(exception, 'error will robinson'):
            self.deliver_email(response_code=code, response_message='error will robinson')

    @override_settings(ACE_CHANNEL_BRAZE_API_KEY='')
    def test_disabled(self):
        assert not self.channel.enabled()
        with self.assertRaisesRegex(FatalChannelDeliveryError, 'disabled'):
            self.deliver_email()



"""
:mod:`edx_ace.channel.sailthru` implements a SailThru-based email delivery
channel for ACE.
"""
import logging
import random
import textwrap
from datetime import datetime, timedelta
from enum import Enum, IntEnum
from gettext import gettext as _

import attr
from dateutil.tz import tzutc

from django.conf import settings

from edx_ace.channel import Channel, ChannelType
from edx_ace.errors import FatalChannelDeliveryError, InvalidMessageError, RecoverableChannelDeliveryError
from edx_ace.utils.date import get_current_time

LOG = logging.getLogger(__name__)

try:
    from sailthru import SailthruClient, SailthruClientError

    CLIENT_LIBRARY_INSTALLED = True
except ImportError:
    LOG.warning('Sailthru client not installed. The Sailthru delivery channel is disabled.')
    CLIENT_LIBRARY_INSTALLED = False


class RecoverableErrorCodes(IntEnum):
    """
    These `error codes`_ are present in responses to requests that can (and should) be retried after waiting for a bit.

    .. _error codes:
        https://getstarted.sailthru.com/developers/api-basics/responses/
    """
    INTERNAL_ERROR = 9
    """
    Something's gone wrong on Sailthru's end. Your request was probably not saved - try waiting a moment and trying
    again.
    """

    RATE_LIMIT = 43
    """
    Too many [type] requests this minute to /[endpoint] API: You have exceeded the limit of requests per minute for the
    given type (GET or POST) and endpoint. For limit details, see the Rate Limiting section on the API Technical Details
    page.
    """


NEXT_ATTEMPT_DELAY_SECONDS = 10

# TODO(later): Should we do something different with these responses?
# OPT_OUT_ERROR_CODES = frozenset([
#     32,  # Email has opted out of delivery from client: This email has opted out of delivery from any emails coming
#          # from your site and should not be emailed.
#     33,  # Email has opted out of delivery from template: This email has opted out of delivery from the specific
#          # template you are sending, and should not be sent this type of email.
#     34,  # Email may not be emailed: This email has been identified as an email that should never be emailed.
#     35,  # Email is a known hardbounce: This email has been previously identified as a hardbounce, so should not be
#          # emailed.
#     37,  # Email will only accept basic templates: The user has opted out of delivery from all templates except basic
#          # templates.
# ])


class ResponseHeaders(Enum):
    """
    These are `special headers`_ returned in responses from the Sailthru REST API.

    .. _special headers:
        https://getstarted.sailthru.com/developers/api-basics/technical/#HTTP_Response_Headers
    """

    RATE_LIMIT_REMAINING = 'X-Rate-Limit-Remaining'
    RATE_LIMIT_RESET = 'X-Rate-Limit-Reset'


class SailthruEmailChannel(Channel):
    """
    An email channel for delivering messages to users using Sailthru.

    This channel makes use of the Sailthru REST API to send messages. It is designed for "at most once" delivery of
    messages. It will make a reasonable attempt to deliver the message and give up if it can't. It also only confirms
    that Sailthru has received the request to send the email, it doesn't actually confirm that it made it to the
    recipient.

    The integration with Sailthru requires several Django settings to be defined.

    Example:

        Sample settings::

            .. settings_start
            ACE_CHANNEL_SAILTHRU_DEBUG = False
            ACE_CHANNEL_SAILTHRU_TEMPLATE_NAME = "Some template name"
            ACE_CHANNEL_SAILTHRU_API_KEY = "1234567890"
            ACE_CHANNEL_SAILTHRU_API_SECRET = "this is secret"
            .. settings_end

    The named template in Sailthru should be minimal, most of the rendering happens within ACE. The "From Name" field
    should be set to ``{{ace_template_from_name}}``. The "Subject" field should be set to ``{{ace_template_subject}}``.
    The "Code" for the template should be::

        <html xmlns="http://www.w3.org/1999/xhtml" lang="en">
            <head>
                {{ace_template_head_html}}
            </head>
            <body>
                {body_html = replace(ace_template_body_html, '{view_url}', view_url)}
                {body_html = replace(body_html, '{optout_confirm_url}', optout_confirm_url)}
                {body_html = replace(body_html, '{forward_url}', forward_url)}
                {body_html = replace(body_html, '{beacon_src}', beacon_src)}
                {body_html}
                <span id="sailthru-message-id" style="display: none;">{message_id()}</span>
                <a href="{optout_confirm_url}" style="display: none;"></a>
            </body>
        </html>

    """

    channel_type = ChannelType.EMAIL

    @classmethod
    def enabled(cls):
        """
        Returns: True iff all required settings are not empty and the Sailthru client library is installed.
        """
        required_settings = (
            'ACE_CHANNEL_SAILTHRU_TEMPLATE_NAME',
            'ACE_CHANNEL_SAILTHRU_API_KEY',
            'ACE_CHANNEL_SAILTHRU_API_SECRET',
        )

        for setting in required_settings:
            if not hasattr(settings, setting):
                LOG.warning('%s is not set, Sailthru email channel is disabled.', setting)

        if not CLIENT_LIBRARY_INSTALLED:
            LOG.warning('The Sailthru API client is not installed, so the Sailthru email channel is disabled.')

        return CLIENT_LIBRARY_INSTALLED and all(
            hasattr(settings, required_setting)
            for required_setting in required_settings
        )

    @property
    def action_links(self):
        """Provides list of action links, called by templates directly"""
        # Note that these variables are evaluated by Sailthru, not the Django template engine
        return [
            ('{view_url}', _('View on Web')),
            ('{optout_confirm_url}', _('Unsubscribe from this list')),
        ]

    @property
    def tracker_image_sources(self):
        """Provides list of trackers, called by templates directly"""
        # Note {beacon_src} is not a template variable that is evaluated by the Django template engine.
        # It is evaluated by Sailthru when the email is sent.
        return ['{beacon_src}']

    def __init__(self):
        if not self.enabled():
            self.sailthru_client = None
        else:
            self.sailthru_client = SailthruClient(
                settings.ACE_CHANNEL_SAILTHRU_API_KEY,
                settings.ACE_CHANNEL_SAILTHRU_API_SECRET,
            )

        self.template_name = settings.ACE_CHANNEL_SAILTHRU_TEMPLATE_NAME

    def deliver(self, message, rendered_message):
        if message.recipient.email_address is None:
            raise InvalidMessageError(
                'No email address specified for recipient {} while sending message {}'.format(
                    message.recipient,
                    message.log_id
                )
            )

        template_vars, options = {}, {}
        for key, value in attr.asdict(rendered_message).items():
            if value is not None:
                # Sailthru will silently fail to send the email if the from name or subject line contain new line
                # characters at the beginning or end of the string
                template_vars['ace_template_' + key] = value.strip()

        if 'reply_to' in message.options and message.options.get('reply_to'):
            options['behalf_email'] = message.options.get('reply_to')
        elif 'from_address' in message.options:
            options['behalf_email'] = message.options.get('from_address')

        logger = message.get_message_specific_logger(LOG)

        if getattr(settings, 'ACE_CHANNEL_SAILTHRU_DEBUG', False):
            logger.info(
                # TODO(later): Do our splunk parsers do the right thing with multi-line log messages like this?
                textwrap.dedent("""\
                    Would have emailed using:
                        template: %s
                        recipient: %s
                        variables: %s
                        options: %s
                """),
                self.template_name,
                message.recipient.email_address,
                str(template_vars),
                str(options),
            )
            return

        if not self.enabled():
            raise FatalChannelDeliveryError(
                textwrap.dedent("""\
                    Sailthru channel is disabled, unable to send:
                        template: %s
                        recipient: %s
                        variables: %s
                        options: %s
                """),
                self.template_name,
                message.recipient.email_address,
                str(template_vars),
                str(options),
            )

        try:
            logger.debug('Sending to Sailthru')

            response = self.sailthru_client.send(
                self.template_name,
                message.recipient.email_address,
                _vars=template_vars,
                options=options,
            )

            if response.is_ok():
                logger.debug('Successfully send to Sailthru')
                # TODO(later): emit some sort of analytics event?
                return
            else:
                logger.debug('Failed to send to Sailthru')
                self._handle_error_response(response)

        except SailthruClientError as exc:
            raise FatalChannelDeliveryError(
                'Unable to communicate with the Sailthru API: ' + str(exc)
            ) from exc  # pragma: no cover

    def _handle_error_response(self, response):
        """
        Handle an error response from SailThru, either by retrying or failing
        with an appropriate exception.

        Arguments:
            response: The HTTP response recieved from SailThru.
        """
        error = response.get_error()
        error_code = error.get_error_code()
        error_message = error.get_message()
        http_status_code = response.get_status_code()
        if error_code in [rec.value for rec in RecoverableErrorCodes]:
            next_attempt_time = None
            if error_code == RecoverableErrorCodes.RATE_LIMIT:
                next_attempt_time = self._get_rate_limit_reset_time(sailthru_response=response)

            if next_attempt_time is None:
                # Sailthru advises waiting "a moment" and then trying again.
                next_attempt_time = get_current_time() + timedelta(
                    seconds=NEXT_ATTEMPT_DELAY_SECONDS + random.uniform(-2, 2)
                )

            raise RecoverableChannelDeliveryError(
                'Recoverable Sailthru error (error_code={error_code} status_code={http_status_code}): '
                '{message}'.format(
                    error_code=error_code,
                    http_status_code=http_status_code,
                    message=error_message
                ),
                next_attempt_time
            )

        raise FatalChannelDeliveryError(
            'Fatal Sailthru error (error_code={error_code} status_code={http_status_code}): '
            '{message}'.format(
                error_code=error_code,
                http_status_code=http_status_code,
                message=error_message
            )
        )

    @staticmethod
    def _get_rate_limit_reset_time(sailthru_response):
        """
        Given a response from the Sailthru API, check to see if our requests are being rate limited.

        If so, return a timestamp indicating when the limit will reset.

        Args:
            sailthru_response (SailthruResponse):

        Returns:
            datetime: The time at which delivery can be re-attempted because the rate limit will be reset.
        """
        response = sailthru_response.response
        headers = response.headers

        try:
            remaining = int(headers[ResponseHeaders.RATE_LIMIT_REMAINING])
            if remaining > 0:
                return None

            reset_timestamp = int(headers[ResponseHeaders.RATE_LIMIT_RESET])
            return datetime.utcfromtimestamp(reset_timestamp).replace(tzinfo=tzutc())
        except ValueError:
            return None



"""
Tests of :mod:`edx_ace.message`.
"""
import logging
from functools import partial
from unittest import TestCase
from unittest.mock import patch

import ddt
from hypothesis import given
from hypothesis import strategies as st
from hypothesis.extra.pytz import timezones

from django.utils.translation import ugettext_lazy

from edx_ace.message import Message, MessageType
from edx_ace.recipient import Recipient
from edx_ace.serialization import MessageEncoder
from edx_ace.utils.date import get_current_time

LOG = logging.getLogger(__name__)

context_values = st.one_of(st.text(), st.booleans(), st.floats(allow_nan=False))
dates = st.datetimes(timezones=st.none() | timezones() | st.none())

msg = st.builds(
    Message,
    app_label=st.text(),
    name=st.text(),
    expiration_time=dates,
    context=st.dictionaries(
        st.text(),
        context_values,
    ),
    recipient=st.builds(
        Recipient,
        lms_user_id=st.integers(min_value=1),
    ),
    send_uuid=st.one_of(st.uuids(), st.none()),
)


@ddt.ddt
class TestMessage(TestCase):
    """
    Tests of :class:`.Message` and :class:`.MessageType`.
    """
    def setUp(self):
        super().setUp()
        self.msg_kwargs = {
            'app_label': 'test_app_label',
            'name': 'test_message',
            'expiration_time': get_current_time(),
            'context': {
                'key1': 'value1',
                'key2': 'value2',
            },
            'recipient': Recipient(
                lms_user_id=123,
            )
        }

        self.encoder = MessageEncoder()

    def test_basic(self):
        transactional_message = Message(options={'transactional': True}, **self.msg_kwargs)
        for key in self.msg_kwargs:
            self.assertEqual(getattr(transactional_message, key), self.msg_kwargs.get(key))
        self.assertIsNotNone(transactional_message.uuid)
        assert transactional_message.options.get('transactional')

        normal_message = Message(**self.msg_kwargs)
        assert not dict(normal_message.options)

    def test_serialization(self):
        message = Message(**self.msg_kwargs)
        string_value = str(message)
        resurrected_msg = Message.from_string(string_value)
        self.assertEqual(message, resurrected_msg)

    def test_serialization_lazy_text(self):
        unicode_text = "A 𝓾𝓷𝓲𝓬𝓸𝓭𝓮 Text"
        lazy_text = ugettext_lazy(unicode_text)                    # pylint: disable=translation-of-non-string
        self.assertEqual(self.encoder.default(lazy_text), unicode_text)

    @given(msg)
    def test_serialization_round_trip(self, message):
        serialized = str(message)
        parsed = Message.from_string(serialized)
        self.assertEqual(message, parsed)

    @ddt.data(
        (None, True, False),
        (logging.WARNING, True, False),
        (logging.DEBUG, True, True),
    )
    @ddt.unpack
    def test_log_level(self, log_level, expect_log_warn, expect_log_debug):
        logging.getLogger().setLevel(logging.INFO)

        self.msg_kwargs['log_level'] = log_level
        message = Message(**self.msg_kwargs)
        logger = message.get_message_specific_logger(LOG)
        with patch('logging.Logger.callHandlers') as mock_log:
            logger.warning('Test warning statement')
            self.assertEqual(mock_log.called, expect_log_warn)
        with patch('logging.Logger.callHandlers') as mock_log:
            logger.debug('Test debug statement')
            self.assertEqual(mock_log.called, expect_log_debug)


def mk_message_type(name, app_label):
    """
    Create a new :class:`MessageType` subclass for testing purposes.

    Arguments:
        name: The NAME class attribute to set.
        app_label: The APP_LABEL class attribute to set.
    """
    class CustomType(MessageType):
        NAME = name
        APP_LABEL = app_label

    return CustomType


msg_type = st.builds(
    mk_message_type,
    name=st.one_of(st.none(), st.text()),
    app_label=st.text(),
).flatmap(
    partial(
        st.builds,
        expiration_time=dates,
        context=st.dictionaries(
            st.text(),
            context_values,
        ),
    )
)


class TestMessageTypes(TestCase):
    """ Test Message Types. """
    @given(msg_type)
    def test_serialization_roundtrip(self, message_type):
        serialized = str(message_type)
        parsed = MessageType.from_string(serialized)
        self.assertEqual(message_type, parsed)
